from django.test import TestCase
from mock import Mock, MagicMock

from care.compat import CursorWrapper

try:
    from django.db.backends.util import CursorWrapper  # noqa
except ImportError:

    # For Django < 1.3

    class CursorWrapperGetAttrTest(TestCase):

        def test_should_proxy_attr(self):
            cursor = Mock()
            db = MagicMock()
            cursor_wrapper = CursorWrapper(cursor, db)

            execute = cursor_wrapper.execute

            self.assertEqual(execute, cursor.execute)

        def test_should_return_cursor_iterator(self):
            cursor = []
            db = MagicMock()
            cursor_wrapper = CursorWrapper(cursor, db)

            old_cursor = iter(cursor_wrapper)

            self.assertEqual(list(old_cursor), cursor)



""" Middlewares. """

from care.models import request_data


class AnnotateSQLMiddleware(object):

    """ Middleware that stores request path in threads local data.

    It stores request path and view name (if exists) for use
    in AnnotatingCursorWrapper, which apends that data to sql queries.

    .. versionadded:: 1.1
    """

    def process_request(self, request):
        """ Store path in threads local data. """
        request_data.path = request.build_absolute_uri()

    def process_view(self, request, view_func, view_args, view_kwargs):
        """ Store view in threads local data. """
        module_name = getattr(view_func, '__module__', '<no module>')
        func_name = getattr(view_func, '__name__', '<no view>')
        request_data.view = '%s.%s' % (module_name, func_name)

    def process_response(self, request, response):
        """ Clear stored path and view. """
        request_data.path = ''
        request_data.view = ''
        return response



from django.test import TestCase
from mock import sentinel
from smarttest.compat import RequestFactory

from care.middleware import AnnotateSQLMiddleware
from care.models import request_data


class AnnotateSQLMiddlewareProcessRequestTest(TestCase):

    def test_should_set_path(self):
        factory = RequestFactory()
        request = factory.get('/test_path')
        middleware = AnnotateSQLMiddleware()

        middleware.process_request(request)
        self.assertEqual(request_data.path, 'http://testserver/test_path')


class AnnotateSQLMiddlewareProcessViewTest(TestCase):

    def test_should_set_view(self):
        factory = RequestFactory()
        request = factory.get('/test_path')
        middleware = AnnotateSQLMiddleware()
        view_func = self.test_should_set_view

        middleware.process_view(request, view_func, (), {})

        self.assertEqual(request_data.view,
                         'care.tests.test_middleware.test_should_set_view')


class AnnotateSQLMiddlewareProcessResponseTest(TestCase):

    def test_should_clear_path_and_view(self):
        middleware = AnnotateSQLMiddleware()

        middleware.process_response(sentinel.request, sentinel.response)

        self.assertEqual(request_data.path, '')
        self.assertEqual(request_data.view, '')



""" Compatibility module. """
try:
    from django.db.backends.util import CursorWrapper
except ImportError:

    # Copied form Django 1.3

    class CursorWrapper(object):

        """ Adds support for CursorWrapper for Django 1.2. """

        def __init__(self, cursor, db):
            """ Create CursorWrapper. """
            self.cursor = cursor
            self.db = db

        def __getattr__(self, attr):
            """ Proxy attributes to cursor. """
            if self.db.is_managed():
                self.db.set_dirty()
            if attr in self.__dict__:
                return self.__dict__[attr]  # pragma: no cover
            else:
                return getattr(self.cursor, attr)

        def __iter__(self):
            """ Create iterator over cursor. """
            return iter(self.cursor)


__all__ = ('CursorWrapper',)



""" Models. """
import threading

from django.conf import settings
from django.db.backends import BaseDatabaseWrapper

from care.compat import CursorWrapper

request_data = threading.local()
request_data.path = ''
request_data.view = ''


class AnnotatingCursorWrapper(CursorWrapper):

    """ Cursor wrapper appending path into sql comment.

    This wrapper appends request path and resolved view into sql comment
    in order to easier find source of query in slow log.
    Wrapper automaticaly monkey-paches BaseDatabaseWrapper
    if AnnotateSQLMiddleware is in MIDDLEWARE_CLASSES.

    .. versionadded:: 1.1
    """

    def __init__(self, cursor_wrapper):
        """ Wrap existing cursor wrapper.

        :param cursor_wrapper: CursorWrapper instance
        """
        self.wrapped = cursor_wrapper

    def execute(self, sql, params=()):
        """ Execute sql query.

        :param sql: sql query
        :param params: sql params to interpolate
        """
        path = getattr(request_data, 'path', '')
        view = getattr(request_data, 'view', '')
        sql += ' -- path: %s | view: %s' % (path.replace('%', '%%'), view.replace('%', '%%'))
        return self.wrapped.execute(sql, params)

    def executemany(self, sql, param_list):
        """ Execute sql query.

        :param sql: sql query
        :param params: sql params to interpolate
        """
        path = getattr(request_data, 'path', '')
        view = getattr(request_data, 'view', '')
        sql += ' -- path: %s | view: %s' % (path.replace('%', '%%'), view.replace('%', '%%'))
        return self.wrapped.executemany(sql, param_list)

    def __getattribute__(self, name):
        """ Proxy attributes to CursorWrapper instance. """
        if name in ['execute', 'executemany', 'wrapped']:
            return object.__getattribute__(self, name)
        return object.__getattribute__(self.wrapped, name)


if 'care.middleware.AnnotateSQLMiddleware' in settings.MIDDLEWARE_CLASSES:
    old_cursor = BaseDatabaseWrapper.cursor
    BaseDatabaseWrapper.cursor = lambda *args: AnnotatingCursorWrapper(old_cursor(*args))



# -*- coding: utf-8 -*-
""" Command that checks which queries are executed on Django's project startup. """

import csv
import logging
from optparse import make_option

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections, reset_queries
from django.utils.importlib import import_module
from six.moves import reload_module

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Writer(object):

    """ Sink type object that stores quereies and output statistics. """

    def __init__(self, stdout, output):
        """ Create Write object.

        :param stdout: File-like object to which save stats
        :param output: file name to which save queries logs
        """
        self.stdout = csv.writer(stdout)
        self.output = open(output, 'w') if output else None
        self.header = ['Module', 'Connection', 'Queries num', 'Total time']
        self.module_name = ''

    def set_module(self, module_name):
        """ Set currently processed module name.

        :param str module_name: module name
        """
        self.module_name = module_name

    def write(self, connection, queries):
        """ Write queries.

        :param connection: connection name
        :param queries: list of queries
        """
        self.write_stats(connection, queries)
        self.write_queries(connection, queries)

    def write_stats(self, connection, queries):
        """ Write queries stats to stdout.

        :param connection: connection name
        :param num: total number of queries
        """
        if self.header:
            self.stdout.writerow(self.header)
            self.header = None
        total_time = sum(float(query['time']) for query in queries)
        total_num = len(queries)
        self.stdout.writerow([self.module_name, connection, total_num, total_time])

    def write_queries(self, connection, queries):
        """ Write queries in log file.

        :param connection: connection name
        :param queries: list of queries
        """
        if self.output is not None:
            line = '%s:%s' % (self.module_name, connection)
            self.output.write('\n%s\n%s\n\n' % (line, '=' * len(line)))
            for i, query in enumerate(queries):
                row = '%s. %s - %s\n' % (i + 1, query['time'], query['sql'])
                self.output.write(row)

    def close(self):
        """ Close log file. """
        if self.output is not None:
            self.output.close()


class Command(BaseCommand):

    """ Command that checks which queries are executed on Django's project startup. """

    option_list = BaseCommand.option_list + (
        make_option('-o', '--output', dest='output', default=None,
                    help="File's name in which store executed queries"),
    )

    def handle(self, *args, **options):
        """ Run checks on given applications. """
        self.run_checks(args, options['output'])

    def run_checks(self, args, output):
        """ Run checks.

        :param args: iterable with application names
        :param output: file name to which save queries
        """
        modules = self.prepare_modules(args)

        writer = Writer(self.stdout, output)
        try:
            for module_name in modules:
                self.check_module(module_name, writer)
        finally:
            writer.close()

    def prepare_modules(self, args):
        """ Prepare list of modules that needs to be imported when checkin queries.

        :param args: iterable with application names
        :returns: list of module names
        """
        apps = args if args else settings.INSTALLED_APPS
        modules = []
        for app in apps:
            modules.append(app)
            for submodule in ['models', 'admin']:
                modules.append('%s.%s' % (app, submodule))
        return modules

    def check_module(self, module_name, writer):
        """ Check module queries.

        :param str module_name: module name
        :param writer: Writer object that will store queries
        """
        if self.reload_module(module_name):
            writer.set_module(module_name)
            self.store_queries(writer)

    def reload_module(self, module_name):
        """ Reload module and reset queries log.

        :param str module_name: module name
        :returns: True on success
        """
        try:
            mod = import_module(module_name)
            reset_queries()
            reload_module(mod)
        except (ImportError, RuntimeError):
            pass
        except Exception as e:
            log.warning(repr(e))
        else:
            return True

    def store_queries(self, writer):
        """ Store queries in writer.

        :param writer: Writer object that will store queries
        """
        for key in connections:
            queries = connections[key].queries
            if queries:
                writer.write(key, queries)



from django.conf import settings
from django.core.management import call_command
from django.test import TestCase
from mock import patch, call, sentinel, MagicMock
import six
from six.moves import StringIO

from smarttest.decorators import no_db_testcase

from care.management.commands.check_startup_queries import Command, Writer

open_builtin = '__builtin__.open' if six.PY2 else 'builtins.open'


@no_db_testcase
class CheckStartupQueriesTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries`
    """

    @patch('care.management.commands.check_startup_queries.reset_queries')
    @patch('care.management.commands.check_startup_queries.import_module')
    @patch('care.management.commands.check_startup_queries.reload_module')
    @patch('care.management.commands.check_startup_queries.connections')
    @patch('care.management.commands.check_startup_queries.Writer')
    def test_should_check_startup_queries(self, Writer, connections, reload_module, import_module, reset_queries):
        out = StringIO()
        err = StringIO()
        apps = [
            'care',
            'django.contrib.auth',
            'django.contrib.contenttypes'
        ]
        connections.__iter__.return_value = ['default', 'slave']
        queries = [{'sql': 'sql', 'time': '0.01'}]
        connection = MagicMock(queries=queries)
        connections.__getitem__.side_effect = [connection] * 2 * len(apps) * 3

        INSTALLED_APPS = list(settings.INSTALLED_APPS)
        settings.INSTALLED_APPS = apps
        try:
            call_command('check_startup_queries', stdout=out, stderr=err)
        finally:
            settings.INSTALLED_APPS = INSTALLED_APPS

        module_names = list(apps)
        module_names += ['%s.models' % app for app in apps]
        module_names += ['%s.admin' % app for app in apps]
        import_moule_calls = [call(module_name) for module_name in module_names]
        import_module.assert_has_calls(import_moule_calls, any_order=True)
        set_module_calls = [call(module_name) for module_name in module_names]
        Writer().set_module.assert_has_calls(set_module_calls, any_order=True)
        Writer().write.assert_has_calls([
            call('default', queries),
            call('slave', queries),
        ])


@no_db_testcase
class CheckStartupQueriesCommandRunChecksTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Command.run_checks`
    """

    @patch.object(Command, 'prepare_modules', spec=Command)
    @patch.object(Command, 'check_module', spec=Command)
    @patch('care.management.commands.check_startup_queries.Writer')
    def test_should_check_startup_queries(self, Writer, check_module, prepare_modules):
        module_names = ['care.models', 'care.admin']
        prepare_modules.side_effect = [module_names]
        command = Command()
        command.stdout = sentinel.stdout

        command.run_checks([], None)

        expected_calls = [call(module_name, Writer())
                          for module_name in module_names]
        check_module.assert_has_calls(expected_calls)


@no_db_testcase
class CheckStartupQueriesCommandPrepareModulesTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Command.prepare_modules`
    """

    def test_should_return_all_applications_modules_list(self):
        command = Command()
        apps = ['care', 'django.contrib.auth']
        INSTALLED_APPS = list(settings.INSTALLED_APPS)
        settings.INSTALLED_APPS = apps
        try:
            modules = command.prepare_modules([])
        finally:
            settings.INSTALLED_APPS = INSTALLED_APPS

        self.assertTrue('care' in modules)
        self.assertTrue('care.models' in modules)
        self.assertTrue('care.admin' in modules)
        self.assertTrue('django.contrib.auth' in modules)
        self.assertTrue('django.contrib.auth.models' in modules)
        self.assertTrue('django.contrib.auth.admin' in modules)

    def test_should_return_application_modules_list(self):
        command = Command()
        app = 'care'
        modules = command.prepare_modules([app])

        self.assertTrue('care' in modules)
        self.assertTrue('care.models' in modules)
        self.assertTrue('care.admin' in modules)


@no_db_testcase
class CheckStartupQueriesCommandCheckModuleTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Command.check_module`
    """

    @patch('care.management.commands.check_startup_queries.Writer')
    @patch.object(Command, 'reload_module', spec=Command)
    @patch.object(Command, 'store_queries', spec=Command)
    def test_should_check_module(self, store_queries, reload_module, Writer):
        module_name = 'care.models'
        writer = Writer()
        command = Command()
        reload_module.side_effect = [True]

        command.check_module(module_name, writer)

        writer.set_module.assert_called_once_with(module_name)
        store_queries.assert_called_once_with(writer)

    @patch('care.management.commands.check_startup_queries.Writer')
    @patch.object(Command, 'reload_module', spec=Command)
    @patch.object(Command, 'store_queries', spec=Command)
    def test_should_not_check_module_if_could_not_reload(self, store_queries, reload_module, Writer):
        module_name = 'care.models'
        writer = Writer()
        command = Command()
        reload_module.side_effect = [False]

        command.check_module(module_name, writer)

        self.assertFalse(writer.set_module.called)
        self.assertFalse(store_queries.called)


@no_db_testcase
class CheckStartupQueriesCommandReloadModuleTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Command.reload_module`
    """

    @patch('care.management.commands.check_startup_queries.reset_queries')
    @patch('care.management.commands.check_startup_queries.import_module')
    @patch('care.management.commands.check_startup_queries.reload_module')
    def test_should_reload_module(self, reload_module, import_module, reset_queries):
        module_name = 'care.models'
        command = Command()

        command.reload_module(module_name)

    @patch('care.management.commands.check_startup_queries.reset_queries')
    @patch('care.management.commands.check_startup_queries.import_module')
    @patch('care.management.commands.check_startup_queries.reload_module')
    def test_should_not_reload_module_if_could_not_import_module(self, reload_module, import_module, reset_queries):
        module_name = 'care.models'
        command = Command()
        import_module.side_effect = [ImportError('%s not found' % module_name)]

        command.reload_module(module_name)

    @patch('care.management.commands.check_startup_queries.reset_queries')
    @patch('care.management.commands.check_startup_queries.import_module')
    @patch('care.management.commands.check_startup_queries.reload_module')
    def test_should_not_reload_module_if_could_not_reload(self, reload_module, import_module, reset_queries):
        module_name = 'care.models'
        command = Command()
        reload_module.side_effect = [Exception('unknown')]

        command.reload_module(module_name)


@no_db_testcase
class CheckStartupQueriesCommandStoreQueriesTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Command.store_queries`
    """

    @patch('care.management.commands.check_startup_queries.connections')
    @patch('care.management.commands.check_startup_queries.Writer')
    def test_should_store_queries(self, Writer, connections):
        writer = Writer()
        connections.__iter__.return_value = ['default', 'slave']
        queries = [{'sql': 'sql', 'time': '0.01'}]
        connection = MagicMock(queries=queries)
        connections.__getitem__.side_effect = [connection] * 2
        command = Command()

        command.store_queries(writer)

        writer.write.assert_has_calls([
            call('default', queries),
            call('slave', queries),
        ])

    @patch('care.management.commands.check_startup_queries.connections')
    @patch('care.management.commands.check_startup_queries.Writer')
    def test_should_not_store_queries_if_not_executed(self, Writer, connections):
        writer = Writer()
        connections.__iter__.return_value = ['default', 'slave']
        queries = []
        connection = MagicMock(queries=queries)
        connections.__getitem__.side_effect = [connection] * 2
        command = Command()

        command.store_queries(writer)

        self.assertFalse(writer.write.called)


@no_db_testcase
class CheckStartupQueriesWriterSetModuleTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Writer.set_module`
    """

    def test_should_store_current_module(self):
        module_name = 'care.models'
        stdout = StringIO()
        writer = Writer(stdout, '')

        writer.set_module(module_name)

        self.assertEqual(writer.module_name, module_name)


@no_db_testcase
class CheckStartupQueriesWriterWriteTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Writer.write`
    """

    @patch.object(Writer, 'write_stats', spec=Writer)
    @patch.object(Writer, 'write_queries', spec=Writer)
    def test_should_write_queries(self, write_queries, write_stats):
        stdout = StringIO()
        writer = Writer(stdout, '')
        queries = [sentinel.query]

        writer.write(sentinel.connection, queries)

        writer.write_stats.assert_called_once_with(
            sentinel.connection, queries)
        writer.write_queries.assert_called_once_with(
            sentinel.connection, queries)


@no_db_testcase
class CheckStartupQueriesWriterWriteStatsTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Writer.write_stats`
    """

    def test_should_write_stats(self):
        module_name = 'care.models'
        stdout = StringIO()
        writer = Writer(stdout, '')
        writer.module_name = module_name
        connection = 'default'
        queries = [
            {'sql': 'select version()', 'time': '0.01'},
            {'sql': 'select version()', 'time': '0.1'},
            {'sql': 'select version()', 'time': '0.5'},
        ]

        writer.write_stats(connection, queries)

        excected = 'Module,Connection,Queries num,Total time\r\ncare.models,default,3,0.61\r\n'
        self.assertEqual(stdout.getvalue(), excected)

    def test_should_write_subsequent_stats(self):
        module_name = 'care.models'
        stdout = StringIO()
        writer = Writer(stdout, '')
        writer.module_name = module_name
        writer.header = None
        connection = 'default'
        queries = [
            {'sql': 'select version()', 'time': '0.01'},
            {'sql': 'select version()', 'time': '0.1'},
            {'sql': 'select version()', 'time': '0.5'},
        ]

        writer.write_stats(connection, queries)

        excected = 'care.models,default,3,0.61\r\n'
        self.assertEqual(stdout.getvalue(), excected)


@no_db_testcase
class CheckStartupQueriesWriterWriteQueriesTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Writer.write_queries`
    """

    @patch(open_builtin)
    def test_should_write_queries(self, open):
        module_name = 'care.models'
        stdout = StringIO()
        writer = Writer(stdout, 'filename')
        writer.module_name = module_name
        connection = 'default'
        queries = [{'sql': 'select version()', 'time': '0.01'}]

        writer.write_queries(connection, queries)

        open().write.assert_called_with('1. 0.01 - select version()\n')

    @patch(open_builtin)
    def test_should_not_write_queries_if_no_log_file(self, open):
        module_name = 'care.models'
        stdout = StringIO()
        writer = Writer(stdout, '')
        writer.module_name = module_name
        connection = 'default'
        queries = [{'sql': 'select version()', 'time': '0.01'}]

        writer.write_queries(connection, queries)

        self.assertFalse(open().write.called)


@no_db_testcase
class CheckStartupQueriesWriterCloseTest(TestCase):

    """
    :py:meth:`care.management.commands.check_startup_queries.Writer.close`
    """

    @patch(open_builtin)
    def test_should_close_output(self, open):
        stdout = StringIO()
        writer = Writer(stdout, 'filename')
        writer.close()

        open().close.assert_called_with()

    @patch(open_builtin)
    def test_should_not_close_closed_output(self, open):
        stdout = StringIO()
        writer = Writer(stdout, '')
        writer.close()

        self.assertFalse(open().close.called)



from setuptools import setup, find_packages

with open('README.rst') as file:
    long_description = file.read()

setup(
    name='django-care',
    version='0.1.0',
    description='Tools which help finding performance issues in Django projects',
    long_description=long_description,
    url="http://kidosoft.pl",
    author="Jakub STOLARSKI (Dryobates)",
    author_email="jakub.stolarski@kidosoft.pl",
    license="beerware",
    keywords="django monitoring performance",
    packages=find_packages('src', exclude=['example*']),
    package_dir={'': 'src'},
    install_requires=[
        'six',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Programming Language :: Python',
        'Topic :: Utilities'
    ],
)



from django.test import TestCase
from mock import Mock, ANY, patch, MagicMock
from six.moves import reload_module

from smarttest.compat import override_settings

import care.models
from care.compat import CursorWrapper
from care.models import request_data, AnnotatingCursorWrapper


class AnnotatingCursorWrapperExecuteTest(TestCase):

    def test_should_annotate_execute(self):
        cursor = Mock()
        db = MagicMock()
        db.is_managed.side_effect = [False]
        db.wrap_database_errors.__exit__.side_effect = [True]
        wrapped_cursor = CursorWrapper(cursor, db)
        cursor_wrapper = AnnotatingCursorWrapper(wrapped_cursor)
        test_sql = 'select 1;'
        request_data.path = 'test_path'
        request_data.view = 'test_view'

        cursor_wrapper.execute(test_sql)

        cursor.execute.assert_called_once_with(test_sql + ' -- path: test_path | view: test_view', ANY)


class AnnotatingCursorWrapperExecuteManyTest(TestCase):

    def test_should_annotate_executemany(self):
        cursor = Mock()
        db = MagicMock()
        db.is_managed.side_effect = [False]
        db.wrap_database_errors.__exit__.side_effect = [True]
        wrapped_cursor = CursorWrapper(cursor, db)
        cursor_wrapper = AnnotatingCursorWrapper(wrapped_cursor)
        test_sql = 'select 1;'
        request_data.path = 'test_path'
        request_data.view = 'test_view'

        cursor_wrapper.executemany(test_sql, [])

        cursor.executemany.assert_called_once_with(test_sql + ' -- path: test_path | view: test_view', ANY)


class AnnotatingCursorWrapperCallProcTest(TestCase):

    def test_should_not_annotate_call_proc(self):
        cursor = Mock()
        db = MagicMock()
        db.is_managed.side_effect = [False]
        db.wrap_database_errors.__exit__.side_effect = [True]
        wrapped_cursor = CursorWrapper(cursor, db)
        cursor_wrapper = AnnotatingCursorWrapper(wrapped_cursor)
        test_sql = 'select 1;'
        request_data.path = 'test_path'
        request_data.view = 'test_view'

        cursor_wrapper.callproc(test_sql, [])

        cursor.callproc.assert_called_once_with(test_sql, ANY)


class AnnotatingCursorWrapperTest(TestCase):

    def test_should_annotate_executemany(self):
        with override_settings(MIDDLEWARE_CLASSES='care.middleware.AnnotateSQLMiddleware'):
            reload_module(care.models)
            with patch('django.db.backends.sqlite3.base.SQLiteCursorWrapper') as SQLiteCursorWrapper:
                from care.models import request_data
                request_data.path = 'test_path'
                request_data.view = 'test_view'
                from django.db import connection
                test_sql = 'select 1;'
                connection.cursor().execute(test_sql)

                SQLiteCursorWrapper().execute.assert_called_once_with(test_sql + ' -- path: test_path | view: test_view', ANY)
        reload_module(care.models)
